/*
The software accompanying this notice (the "Software") is provided
to the user to facilitate the use of the data collected by the Mars
Science Laboratory MAHLI/MARDI/Mastcam (the "MMM Data").  Malin Space
Science Systems ("MSSS") grants the user (either as an individual or
entity) a personal, non-transferable, and non-exclusive right (i) to
use and reproduce the Software solely for the purpose of accessing the
MMM Data; (ii) to modify the source code of the Software as necessary
to maintain or adapt the Software to run on alternate computer
platforms; and (iii) to compile, use and reproduce the modified versions
of the Software solely for the purpose of accessing the MMM Data.  The
user may not redistribute the Software or any modifications thereof, for
any purpose.

The user acknowledges that no title to the intellectual property in
the Software is transferred to the user.  The user further acknowledges
that title and full ownership rights to the Software will remain the
exclusive property of MSSS or its suppliers, and the user will not
acquire any rights to the Software except as expressly set forth above.
The Software is provided to the user AS IS.  MSSS MAKES NO WARRANTY,
EXPRESS OR IMPLIED, WITH RESPECT TO THE SOFTWARE, AND SPECIFICALLY
DISCLAIMS THE IMPLIED WARRANTIES OF NON-INFRINGEMENT OF THIRD PARTY
RIGHTS, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  SOME
JURISDICTIONS DO NOT ALLOW THE EXCLUSION OR LIMITATION OF INCIDENTAL
OR CONSEQUENTIAL DAMAGES, SO SUCH LIMITATIONS OR EXCLUSIONS MAY NOT
APPLY TO THE USER.

The user's use or reproduction of the Software constitutes the user's
agreement to the terms of this Notice.  If the user does not agree
with the terms of this notice, promptly return or destroy all copies
of the Software in the user's possession.

Copyright (C) 2013, 2014 Malin Space Science Systems.  All Rights Reserved.

With contributions from:
  Tobias Thierer


*/
/*
   dat2img.c

   Process DAT files into IMG format with detached label
   
   usage: dat2img <dat-filename>

   Output files are placed in current directory.

*/

#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <fcntl.h>
#include <stdlib.h>
#include <jpeglib.h>
#include <sys/types.h>
#include <unistd.h>
#include <arpa/inet.h>

int pdecom(int fd, int width, int height, unsigned char *outbuf);

#define HEADER_MAGIC0  0xff00f0ca
#define HEADER_MAGIC1  0x1010cc28
#define JPEG_SOI       0xffd8

#define MISSING_CONSTANT 0

typedef enum {JPEG, PRED, RAW} image_t;

typedef struct {
   int width;
   int height;
   int col_offset;
   int row_offset;
   int bits_per_element;
   int depth;
   int bands;
   int thumbnail;
   image_t image_type;
} header_t;

void print_usage()
{
   printf("Usage: dat2img [-d ] input.dat [output_dir]\n");
   printf("\tUse -d option for detached label file (default is attached label)\n");
   printf("\tDefault ouput file is input_nn.img in the current directory\n");
}

int read_header(int fd, unsigned int *header)
{
   /* Search for magic number in first 1000 bytes */
   
   int offset;
   for (offset = 0; offset < 1000 ; offset++) {
      lseek(fd, offset, SEEK_SET);
      if (read(fd, header, 64) == 0) {
         return -1;
      }
      if (htonl(header[15]) == HEADER_MAGIC1) {
         int i;
         for (i = 0; i < 16; i++)
            header[i] = htonl(header[i]);
         return 0;
      }
   }
   return -1;
}

int decode_dat_header(int fd, header_t *hdr)
{
   unsigned int raw_hdr[16];
   if (read_header(fd, raw_hdr) != 0) {
      close(fd);
      return -1;
   }
   hdr->width = ((raw_hdr[5]>>8) & 0xff)*8;
   if (hdr->width == 0) hdr->width = 1648;
   hdr->height = ((raw_hdr[5]) & 0xff)*8;
   if (hdr->height == 0) hdr->height = 1200;
   hdr->col_offset = ((raw_hdr[5]>>24) & 0xff)*8;
   hdr->row_offset = ((raw_hdr[5]>>16) & 0xff)*8;

   if ((raw_hdr[8] & 0xff) == 0) {
      if (raw_hdr[8] & 0xff00)
    	   hdr->image_type = PRED;
      else
         hdr->image_type = RAW;
   } else {
      hdr->image_type = JPEG;
   }
   hdr->thumbnail = (raw_hdr[0]>>24) & 0x8;

   hdr->bands = 1;
   if (hdr->image_type != JPEG && raw_hdr[9] == 255) {
      hdr->bits_per_element = 16;
      hdr->depth = 2;
   } else if (hdr->image_type == JPEG && (raw_hdr[8] & 0xff00)) {
      hdr->bits_per_element = 8;
      hdr->depth = 1;
      hdr->bands = 3;
   } else {
      hdr->bits_per_element = 8;
      hdr->depth = 1;
   }
   return 0;
}

int pds_newline(int fd, int n)
{
   int nb = 0;
   int cnt = 0;
   for(cnt=0; cnt < n; cnt++)
      nb += write(fd, "\r\n", 2);
   return nb;
}

int pds_write(int fd, char* tag, char *val)
{
   static char line[1024];
   sprintf(line, "%-36s= %s\r\n", tag, val);
   return write(fd, line, strlen(line));
}

int pds_write_int(int fd, char* tag, int val)
{
   static char line[1024];
   sprintf(line, "%-36s= %d\r\n", tag, val);
   return write(fd, line, strlen(line));
}

int pds_comment(int fd, char* comment)
{
   static char line[1024];
   sprintf(line, "\r\n/* %s */\r\n\r\n", comment);
   return write(fd, line, strlen(line));
}

int pds_write_image_name (int fd, char* image_filename)
{
   static char line[1024];
   sprintf(line, "^IMAGE = (\"%s\")\r\n", image_filename);
   return write(fd, line, strlen(line));
}

void write_img_label(int fd, header_t *dat_hdr, int detached, char *image_filename, int record_bytes)
{
   int nbytes = dat_hdr->depth * dat_hdr->width * dat_hdr->height;
   int image_records = (nbytes + record_bytes-1)/record_bytes;
   int label_records = 1;
   int actual_records;
   int file_records = image_records;
   if (!detached) file_records += label_records;

   // loop until the number of records is stable 
   while (1) {
      nbytes = 0;
      nbytes += pds_write(fd,      "PDS_VERSION_ID", "PDS3");
      nbytes += pds_comment(fd,    "FILE DATA ELEMENTS");
      nbytes += pds_write(fd,      "RECORD_TYPE", "FIXED_LENGTH");
      nbytes += pds_write_int(fd,  "RECORD_BYTES", record_bytes);
      nbytes += pds_write_int(fd,  "FILE_RECORDS", file_records);
      nbytes += pds_write_int(fd,  "LABEL_RECORDS", label_records);
      nbytes += pds_comment(fd,    "Pointers to Data Objects");
      if (detached) {
         // Only write the filename, not the path
         char *file_name = strrchr(image_filename, '/');
         if (file_name == NULL) file_name = image_filename;
         else file_name++;

         nbytes += pds_write_image_name (fd, file_name);
      }
      else
         nbytes += pds_write_int(fd, "^IMAGE", label_records+1);
      nbytes += pds_comment(fd,    "IMAGE DATA ELEMENTS");
      nbytes += pds_write(fd,      "OBJECT", "IMAGE");
      nbytes += pds_write(fd,      "  INTERCHANGE_FORMAT", "BINARY");
      nbytes += pds_write_int(fd,  "  LINES", dat_hdr->height);
      nbytes += pds_write_int(fd,  "  LINE_SAMPLES", dat_hdr->width);
      nbytes += pds_write(fd,      "  SAMPLE_TYPE", "UNSIGNED_INTEGER");
      nbytes += pds_write_int(fd,  "  SAMPLE_BITS", dat_hdr->bits_per_element);
      nbytes += pds_write_int(fd,  "  BANDS", dat_hdr->bands);
      nbytes += pds_write(fd,      "  BAND_STORAGE_TYPE", "BAND_SEQUENTIAL");
      nbytes += pds_write(fd,      "  CHECKSUM", "\"N/A\"");
      nbytes += pds_write_int(fd,  "  FIRST_LINE", dat_hdr->row_offset + 1);
      nbytes += pds_write_int(fd,  "  FIRST_LINE_SAMPLE", dat_hdr->col_offset + 1);
      nbytes += pds_write(fd,      "  INVALID_CONSTANT", "0.0");
      nbytes += pds_write_int(fd,  "  MISSING_CONSTANT", MISSING_CONSTANT);
      nbytes += pds_write(fd,      "  SAMPLE_BIT_MASK", "2#11111111#");
      nbytes += pds_write(fd,      "END_OBJECT", "IMAGE");
      nbytes += pds_newline(fd, 3);
      nbytes += write(fd, "END", 3);
      nbytes += pds_newline(fd, 3);

      actual_records = (nbytes + record_bytes-1)/record_bytes;
      if (actual_records != label_records) {
         label_records = actual_records;
         lseek(fd, 0, SEEK_SET);
      } else {
         break;
      }
   }
   if (!detached) {
      // pad to a complete record
      int label_bytes = label_records * record_bytes;
      while (nbytes < label_bytes) nbytes += write(fd, " ", 1); 
   }
}

int write_output_files(char * infile, header_t *dat_hdr, int detached, char * outdir, int nimage, unsigned char * outbuf, int bytes_read)
{
   // determine output file names
   char lbl_filename[1024];
   char img_filename[1024];
   char *base = strrchr(infile, '/');
   if (base == NULL) base = infile;
   else base++;
   strcpy(lbl_filename, outdir);
   strncat(lbl_filename, base, strlen(base)-4);
   lbl_filename[strlen(outdir) + strlen(base)-4] = '\0';
   strcpy(img_filename, lbl_filename);
   char simg[10];
   sprintf(simg, "_%02d.", nimage);
   int lbl_fd, img_fd;
   
   // create image file
   strcat(img_filename, simg);
   strcat(img_filename, "IMG");
   printf("Image file name is: %s\n", img_filename);
   img_fd = open(img_filename, O_CREAT | O_WRONLY | O_TRUNC, 0666);
   if (img_fd < 0) {
        fprintf(stderr, "dat2img: Error: unable to create output IMG file %s\n", img_filename);
        return 1;
   }

   // create label file
   if (detached) {
      strcat(lbl_filename, simg);
      strcat(lbl_filename, "LBL");
      printf("Detached label file name is: %s\n", lbl_filename);
      lbl_fd = open(lbl_filename, O_CREAT | O_WRONLY | O_TRUNC, 0666);
      if (lbl_fd < 0) {
         fprintf(stderr, "dat2img: Error: unable to create output label file %s\n", lbl_filename);
         return 1;
      }
   } else {
      printf("Creating attached label\n");
      lbl_fd = img_fd;
   }

   // write label
   int record_bytes = dat_hdr->width * dat_hdr->depth;
   write_img_label(lbl_fd, dat_hdr, detached, img_filename, record_bytes);
   if (detached) close(lbl_fd);

   // write image data
   write(img_fd, outbuf, bytes_read);

   // pad to a complete record
   int image_records = (bytes_read + record_bytes-1)/record_bytes;
   while (bytes_read < record_bytes * image_records) bytes_read += write(img_fd, "\0", 1); 
   close (img_fd);
   return 0;
}

int jpeg_decom(FILE *fp, int width, int height, int *in_bands, unsigned char *outbuf, int *more_imgs)
{
   struct jpeg_decompress_struct cinfo;
   struct jpeg_error_mgr jerr;
   int bands = *in_bands;

   int save_pos = ftell(fp); // save beginning of image

   cinfo.err = jpeg_std_error(&jerr);
   jpeg_create_decompress(&cinfo);
   jpeg_stdio_src(&cinfo, fp);
   jpeg_read_header(&cinfo, TRUE);
   if (cinfo.num_components != bands) {
      fprintf(stderr, "dat2img: Warning: Mismatch between number of bands in header and in JPEG.\n");
      bands = *in_bands = cinfo.num_components;
   }
   
   jpeg_start_decompress(&cinfo);

   int row_len = cinfo.output_width * cinfo.num_components;

   unsigned char *row_ptr[1];
   row_ptr[0] = (unsigned char *)malloc(row_len);

   int i = 0;
   int j, k;
   // copy output into band interleaved buffers (row_ptr is pixel interleaved)
   unsigned char * out_ptr[bands];
   for(k = 0; k < bands; k++) out_ptr[k] = outbuf + k * width * height;
   while (cinfo.output_scanline < cinfo.output_height) {
      jpeg_read_scanlines(&cinfo, row_ptr, 1);
      for (j = 0; j < row_len; j++)
         *(out_ptr[j % bands]++) = row_ptr[0][j];
      i += cinfo.output_width*cinfo.num_components;
   }
   jpeg_finish_decompress(&cinfo);
   jpeg_destroy_decompress(&cinfo);
   free(row_ptr[0]);

   // look for more images
   unsigned short z;
   *more_imgs = 0;
   fseek(fp, save_pos + 2, SEEK_SET);

   while (fread(&z, 2, 1, fp))
      if (htons(z) == JPEG_SOI) {
         fseek(fp, -2, SEEK_CUR);
         *more_imgs = 1;
         break;
      }
   return i;
}

int set_directory(char *dirname, const char *arg)
{
   strcpy(dirname, arg);
   DIR *dir;
   if ((dir = opendir(dirname))) {
      closedir(dir); 
      if (dirname[strlen(dirname)-1] != '/') // make sure there is a slash at the end
         strcat(dirname, "/");
      return 0; // success
   } else {
      return 1; // directory doesn't exist, or can't be opened
   }
}

int main(int argc, char **argv)
{
   char infile[1024]; // input filename (must be provided by user)
   char outdir[1024]; // output directory (use current if not provided)
   outdir[0] = '\0';
   int detached = 0; // default is attached header
   
   // process arguments
   int idx = 1;
   if (argc < 2 || argc > 4) {
      print_usage();
      return 1;
   }
   // detached or attached header file
   if (strcmp(argv[idx], "-d") == 0) {
      if (argc < 3) {
         print_usage();
         return 1;
      }
      detached = 1; 
      idx++;
   } else if (argv[1][0] == '-') {
      print_usage();
      return 1;
   }

   // input dat file name
   strcpy(infile, argv[idx++]);
   if (strlen(infile) < 4 || strcasecmp(infile+strlen(infile)-4, ".dat")) {
      fprintf(stderr, "dat2img: Error: filename must end with .dat\n");
      return 1;
   }
   // output directory
   if (idx < argc) {
      if (set_directory(outdir, argv[idx++]) != 0) {
         fprintf(stderr, "dat2img: Error: %s is not a existing directory.\n", outdir);
         return 1;
      }
   }

   // process DAT file
   int fd = open(infile, O_RDONLY);
   if (fd < 0) {
      fprintf(stderr, "dat2img: Error: unable to open input file\n");
      return 1;
   }

   // decode 64-byte header from dat file
   header_t dat_hdr;
   if (decode_dat_header(fd, &dat_hdr) < 0) {
      close(fd);
      fprintf(stderr, "dat2img: Error: unable to decode header for %s\n", infile);
      return 1;
   }

   fprintf(stdout, "width = %d, height = %d, bits per pixel = %d\n", 
           dat_hdr.width, dat_hdr.height, dat_hdr.bits_per_element);
   fprintf(stdout, "col_offset = %d, row_offset = %d\n",
           dat_hdr.col_offset, dat_hdr.row_offset);

   // allocate image buffer.  If multiple images in JPEG, they will all have 
   // the same size  as specified by the header
   int image_bytes = dat_hdr.width * dat_hdr.height * dat_hdr.depth * dat_hdr.bands;
   int max_image_bytes = image_bytes;
   if (dat_hdr.image_type == RAW && dat_hdr.thumbnail) // maximum image size may be greater
      max_image_bytes = (dat_hdr.width+7) * (dat_hdr.height+7) * dat_hdr.depth * dat_hdr.bands;
   unsigned char *outbuf = malloc(max_image_bytes);

   FILE *fp = fdopen(fd, "r");
   
   int bytes_read = 0;
   int nimage = 0;
   int more_imgs = 1;
   while (more_imgs) {
      switch(dat_hdr.image_type) {
      case PRED:
         printf("Extracting losslessly compressed (PRED) image...\n");
         bytes_read = pdecom(fd, dat_hdr.width, dat_hdr.height, outbuf);
         more_imgs = 0; // always a single image in a pred file
         break;
      case JPEG:
         printf("Extracting JPEG image %d ...\n", nimage);
         bytes_read = jpeg_decom(fp, dat_hdr.width, dat_hdr.height, &dat_hdr.bands, outbuf, &more_imgs);
         // number of bands may have changed
         image_bytes = dat_hdr.width * dat_hdr.height * dat_hdr.depth * dat_hdr.bands;
         break;
      case RAW:
         printf("Reading in uncompressed image...\n");
         bytes_read = read(fd, outbuf, max_image_bytes);
         if (dat_hdr.depth == 2) {
            // neet to swap bytes
            int i;
            for (i = 0; i < bytes_read; i+=2) {
               char tmp = outbuf[i];
               outbuf[i] = outbuf[i+1];
               outbuf[i+1] = tmp;
            }
         }
         more_imgs = 0; // there is always a single image in raw image file
         //
         // If the image is a raw thumbnail, the height and width in the header may
         // not match the actual image (since the stored values have been divided by eight).  
         // Try to find height and width that match the data.
         // 
         printf("Bytes read = %d\n", bytes_read);

         if (bytes_read !=  image_bytes && dat_hdr.thumbnail)  {
         //
         // Another complication.  There is extra padding in the IMG to a complete record 
         // size (64 bytes). These are the two cases we have found so far.
         //
            int test_1 = bytes_read - 32;
            int test_2 = bytes_read - 56;
            int h, w;
            int fixed = 0;
            for (h = dat_hdr.height; h < dat_hdr.height+8; h++) {
              for (w = dat_hdr.width; w < dat_hdr.width+8; w++) {
                int check_size = h*w*dat_hdr.depth;
                if (check_size == test_1 || check_size == test_2) {
                  dat_hdr.height = h;
                  dat_hdr.width = w;
                  image_bytes = check_size;
                  fixed = 1;
                  fprintf(stdout, "Found actual width = %d, actual height = %d\n",  w, h);
                  break; 
                }
              }
            }
            if (!fixed) {
               fprintf(stderr, "dat2img: Warning: unable to determine image size, assuming 206 x 150.\n");
               dat_hdr.height = 150;
               dat_hdr.width = 206;
               image_bytes = 150 * 206;
            }
          }
          break;
      }

      if (bytes_read < 0)
         fprintf(stderr, "dat2img: Error: unable to decompress image\n");

      if (bytes_read < image_bytes) {
         printf("Warning: truncated input file!!\n\n"); 
         int pos;
         for (pos = bytes_read; pos < image_bytes; pos++) {
            outbuf[pos] = MISSING_CONSTANT;
         }
      }

      write_output_files(infile, &dat_hdr, detached, outdir, nimage, outbuf, image_bytes);
      nimage++;
   }
   fclose(fp); // also closes fd
   free(outbuf);

   return 0;
}
