/*
The software accompanying this notice (the "Software") is provided
to the user to facilitate the use of the data collected by the Mars
Science Laboratory MAHLI/MARDI/Mastcam (the "MMM Data").  Malin Space
Science Systems ("MSSS") grants the user (either as an individual or
entity) a personal, non-transferable, and non-exclusive right (i) to
use and reproduce the Software solely for the purpose of accessing the
MMM Data; (ii) to modify the source code of the Software as necessary
to maintain or adapt the Software to run on alternate computer
platforms; and (iii) to compile, use and reproduce the modified versions
of the Software solely for the purpose of accessing the MMM Data.  The
user may not redistribute the Software or any modifications thereof, for
any purpose.

The user acknowledges that no title to the intellectual property in
the Software is transferred to the user.  The user further acknowledges
that title and full ownership rights to the Software will remain the
exclusive property of MSSS or its suppliers, and the user will not
acquire any rights to the Software except as expressly set forth above.
The Software is provided to the user AS IS.  MSSS MAKES NO WARRANTY,
EXPRESS OR IMPLIED, WITH RESPECT TO THE SOFTWARE, AND SPECIFICALLY
DISCLAIMS THE IMPLIED WARRANTIES OF NON-INFRINGEMENT OF THIRD PARTY
RIGHTS, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  SOME
JURISDICTIONS DO NOT ALLOW THE EXCLUSION OR LIMITATION OF INCIDENTAL
OR CONSEQUENTIAL DAMAGES, SO SUCH LIMITATIONS OR EXCLUSIONS MAY NOT
APPLY TO THE USER.

The user's use or reproduction of the Software constitutes the user's
agreement to the terms of this Notice.  If the user does not agree
with the terms of this notice, promptly return or destroy all copies
of the Software in the user's possession.

Copyright (C) 2013 Malin Space Science Systems.  All Rights Reserved.
*/

/*
   MSL DEA ground software, predictive decompression
   Michael Caplinger, Malin Space Science Systems

   This is a highly preliminary version of the predictive decompressor,
   without much error checking or resync capability, for initial
   FSW testing.  Use at your own risk.

   Compile with "gcc -o pdecom-msl pdecom-msl.c"
   Tested on Solaris 8/SPARC and RHEL4 Intel 64-bit.
*/

#include <stdio.h>   // fgetc(), fprintf(), printf()
#include <stdlib.h>  // exit()
#include <unistd.h>  // dup()
#include <string.h>  // memcpy(), strerror()
#include <errno.h>   // errno

static int debug = 0;

typedef struct {
    unsigned int p;
    unsigned char byte;
    int bit;
    off_t length; // file length
} InputBits_t;

static void bitsInit(int fd, InputBits_t *s)
{
    s->length = lseek( fd, 0, SEEK_END ); // store the file length
    lseek( fd, 0, SEEK_SET ); // and reset to the beginning of the file
    s->p = 0;
    s->bit = 0;
    read(fd, &s->byte, 1);
}

/* return the next bit from the input */
static int nextBit(int fd, InputBits_t *s)
{
   unsigned char cbit;
    
   cbit = (s->byte)>>(7-s->bit);
   cbit &= 0x1;
   s->bit += 1;
   if(s->bit == 8) {
      s->p += 1;
      s->bit = 0;
      read(fd, &s->byte, 1);
      if(debug) fprintf(stderr, "n(%x) ", (s->p));
   }
   if(debug) fprintf(stderr, "%d ", cbit);
   return cbit;
}

/* back up by x bytes and reset s */
static void rewindBytes( int fd, InputBits_t *s, int count )
{
    lseek( fd, -count, SEEK_CUR );
	s->p -= count;
	s->bit = 0;
	read(fd, &s->byte, 1);
}

/*
    These tables represent the Huffman tree.  Each i is a node.  If the
    LEFT bit is set in flags, then left is the index of the next node in
    the tree; if it's clear, left is a leaf and contains the decoded value.
    RIGHT is handled similarly.
*/
static unsigned char data[256*3] = {
0x03,0x03,0x03,0x01,0x03,0x01,0x01,0x03,0x02,0x02,0x00,0x01,0x01,0x03,0x01,0x00,
0x02,0x01,0x01,0x00,0x02,0x02,0x00,0x00,0x03,0x03,0x00,0x03,0x01,0x01,0x03,0x02,
0x02,0x00,0x01,0x01,0x03,0x00,0x01,0x02,0x02,0x00,0x02,0x02,0x00,0x00,0x03,0x03,
0x03,0x01,0x01,0x01,0x03,0x02,0x02,0x00,0x03,0x03,0x03,0x02,0x03,0x03,0x00,0x00,
0x03,0x00,0x00,0x03,0x03,0x03,0x00,0x00,0x03,0x00,0x00,0x03,0x03,0x00,0x00,0x03,
0x00,0x00,0x03,0x03,0x03,0x03,0x00,0x00,0x03,0x00,0x00,0x03,0x03,0x00,0x00,0x03,
0x00,0x00,0x03,0x03,0x03,0x00,0x00,0x03,0x00,0x00,0x03,0x03,0x00,0x00,0x03,0x00,
0x00,0x03,0x03,0x03,0x03,0x03,0x00,0x00,0x03,0x00,0x00,0x03,0x03,0x00,0x00,0x03,
0x00,0x00,0x03,0x03,0x03,0x00,0x00,0x03,0x00,0x00,0x03,0x03,0x00,0x00,0x03,0x00,
0x00,0x03,0x03,0x03,0x03,0x00,0x00,0x03,0x00,0x00,0x03,0x03,0x00,0x00,0x03,0x00,
0x00,0x03,0x03,0x03,0x00,0x00,0x03,0x00,0x00,0x03,0x03,0x00,0x00,0x03,0x00,0x00,
0x02,0x02,0x02,0x01,0x01,0x03,0x03,0x03,0x03,0x03,0x00,0x00,0x03,0x00,0x00,0x03,
0x03,0x00,0x00,0x03,0x00,0x00,0x03,0x03,0x03,0x00,0x00,0x03,0x00,0x00,0x03,0x03,
0x00,0x00,0x03,0x00,0x00,0x03,0x03,0x03,0x03,0x00,0x00,0x03,0x00,0x00,0x03,0x03,
0x00,0x00,0x01,0x00,0x03,0x00,0x01,0x00,0x00,0x03,0x03,0x01,0x01,0x03,0x02,0x02,
0x00,0x01,0x01,0x03,0x00,0x03,0x00,0x03,0x02,0x00,0x00,0x02,0x02,0x00,0x00,0x01,
0x02,0x03,0x04,0x05,0x06,0x07,0x08,0xed,0x17,0x1b,0x0c,0x0d,0x0e,0x0f,0xdd,0x1f,
0x12,0x13,0x2a,0x08,0xf4,0xf0,0xff,0x19,0x1a,0xfb,0x1c,0x1d,0x1e,0x1f,0x14,0x18,
0x1c,0x23,0x24,0x25,0x20,0x27,0x24,0x28,0xaa,0x09,0xf3,0xef,0x02,0x2f,0x30,0x31,
0x32,0x33,0x34,0x35,0x15,0xe7,0x1d,0x39,0x3a,0x3b,0x21,0x3d,0x3e,0xa8,0xa5,0x41,
0xa1,0x9f,0x44,0x45,0x46,0xab,0x9b,0x49,0x9d,0x97,0x4c,0x4d,0x95,0x93,0x50,0x91,
0x8f,0x53,0x54,0x55,0x56,0x9c,0x8b,0x59,0x89,0x87,0x5c,0x5d,0x8d,0x83,0x60,0x81,
0xc3,0x63,0x64,0x65,0x7d,0xc5,0x68,0x79,0x77,0x6b,0x6c,0x75,0xc8,0x6f,0x71,0x6f,
0x72,0x73,0x74,0x75,0x76,0x6d,0x86,0x79,0x69,0x67,0x7c,0x7d,0xbc,0xb5,0x80,0xad,
0x85,0x83,0x84,0x85,0x5d,0x5b,0x88,0x59,0x57,0x8b,0x8c,0x55,0x53,0x8f,0x51,0x4f,
0x92,0x93,0x94,0x95,0x4d,0x4b,0x98,0x49,0x47,0x9b,0x9c,0x45,0x43,0x9f,0x41,0x3f,
0xa2,0xa3,0xa4,0x3d,0x3b,0xa7,0x39,0x37,0xaa,0xab,0x35,0x33,0xae,0x31,0x2f,0x06,
0x0a,0xf2,0xb4,0xb5,0xb6,0xb7,0xb8,0xb9,0xba,0x2d,0x6b,0xbd,0xaf,0xb1,0xc0,0xc1,
0xb7,0xb9,0xc4,0xbe,0xc0,0xc7,0xc8,0xc9,0xca,0xcc,0xcc,0xce,0xd0,0xcf,0xd0,0xd2,
0xd4,0xd3,0xb3,0x61,0xd6,0xd7,0xd8,0xd9,0xbb,0x63,0xdc,0xc1,0x64,0xdf,0xe0,0x73,
0x7a,0xe3,0x7f,0xe5,0x25,0xe7,0xd7,0xfd,0xea,0xeb,0xec,0xed,0xee,0x12,0x16,0x1a,
0xf2,0xf3,0xf4,0x1e,0xf6,0xde,0xf8,0x26,0x2b,0xd9,0x07,0x0b,0xf1,0x04,0x2e,0x18,
0x17,0x00,0x14,0xf8,0x0c,0x0b,0x09,0x0a,0xe5,0xec,0xe8,0x10,0xe1,0x23,0x11,0xdc,
0xd8,0xa9,0x15,0x16,0x10,0x01,0x2d,0x1b,0x05,0x2a,0xf7,0x0d,0x22,0x20,0x21,0xe4,
0xeb,0x19,0x26,0xe0,0xdf,0x28,0x29,0xa6,0x2b,0x2c,0x11,0xfe,0xe9,0xe8,0xb0,0xfa,
0xf6,0x0e,0x38,0x36,0x37,0xe3,0x71,0x52,0x43,0x3c,0x40,0x3f,0xa7,0xa2,0x42,0xa0,
0x9e,0x4b,0x48,0x47,0xa4,0xa3,0x4a,0x98,0x96,0x4f,0x4e,0x94,0x92,0x51,0x90,0x8e,
0x62,0x5b,0x58,0x57,0x9a,0x8a,0x5a,0x88,0x99,0x5f,0x5e,0x84,0x82,0x61,0xc2,0xc4,
0x6a,0x67,0x66,0x7c,0xc6,0x69,0x78,0x76,0x6e,0x6d,0xc7,0x72,0x70,0x70,0x6e,0x91,
0x82,0x7b,0x78,0x77,0x8c,0x6a,0x7a,0x68,0x66,0x7f,0x7e,0xb4,0xac,0x81,0xae,0x6c,
0x8a,0x87,0x86,0x5c,0x5a,0x89,0x58,0x56,0x8e,0x8d,0x54,0x52,0x90,0x50,0x4e,0xa1,
0x9a,0x97,0x96,0x4c,0x4a,0x99,0x48,0x46,0x9e,0x9d,0x44,0x42,0xa0,0x40,0x3e,0xa9,
0xa6,0xa5,0x3c,0x3a,0xa8,0x38,0x36,0xad,0xac,0x34,0x32,0xaf,0x30,0x2e,0xb1,0xb2,
0xb3,0xee,0xea,0xd5,0xc6,0xbf,0xbc,0xbb,0x2c,0x5f,0xbe,0xb0,0xb2,0xc3,0xc2,0xb8,
0xba,0xc5,0xbf,0xc9,0xce,0xcb,0xca,0xcb,0xcd,0xcd,0xcf,0xd1,0xd2,0xd1,0xd3,0x5e,
0xd4,0xb6,0x60,0xe4,0xde,0xdb,0xda,0xbd,0x62,0xdd,0x65,0x74,0xe2,0xe1,0x7b,0x7e,
0x29,0x80,0xe6,0xdb,0xda,0xd6,0x03,0xfe,0xfb,0xf9,0xf5,0xf1,0xef,0xf0,0xe6,0x13,
0xe9,0xf5,0xe2,0xf7,0x22,0xfa,0xf9,0xd5,0x27,0xfc,0xfd,0x0f,0xfc};

static unsigned char *flags = &data[0];

static unsigned char *left = &data[255];

static unsigned char *right = &data[255*2];

#define LEFT 1
#define RIGHT 2

static int nextValue(int fd, InputBits_t *s)
{
   int node = 0; /* start at the root */
   int bit;

   while(1) {
      bit = nextBit(fd, s);
#ifndef NO_EOF
      if(bit == -1) return -1;
#endif
      if(bit == 0) {
         /* go left */
         if(flags[node]&LEFT) node = left[node];
         else return left[node];
      } else {
         /* go right */
         if(flags[node]&RIGHT) node = right[node];
         else return right[node];
      }
   }
}

#define PREDSYNC 0xffff0000
#define true 1
#define false 0

static int decodeX(int fd, int width, unsigned char *out, InputBits_t *s)
{
   unsigned char prev = 0; /* previous pixel is treated as zero */
   int val, i;
   unsigned char temp;
   unsigned long res;
   unsigned int sync = 0;

    lseek(fd, 0, SEEK_CUR);
    long nextSync = -1;

    do
    {
        sync = 0;
        for(i = 0; i < 32; i++)
        {
            sync |= nextBit(fd, s)<<(31-i);
        }
        if( lseek(fd, 0, SEEK_CUR) >= s->length )
        {
            if(debug) fprintf(stderr, "Found end of file instead of sync\n" );
            return 1;
        }
        if( PREDSYNC != sync )
        {
            // back up three bytes, reset and try again to find the sync
            long pos = lseek(fd, 0, SEEK_CUR);
            if( pos != -1 )
            {
                rewindBytes( fd, s, 4 );
                if( debug && (0 == (pos % 100)) )
                    fprintf( stderr, "bad sync %08x at offset %ld, should be 0x%08x\n", sync, pos, PREDSYNC );
            }
        }
    }
    while( PREDSYNC != sync );

    if(sync != PREDSYNC )
    {
        long pos = lseek(fd, 0, SEEK_CUR);
        if ( pos == -1 )
        {
            // stdin cannot be seeked()
            fprintf( stderr, "bad sync %08x, should be 0x%08x\n", sync, PREDSYNC );
        }
        else
        {
            fprintf( stderr, "bad sync %08x at offset %ld, should be 0x%08x\n", sync, pos, PREDSYNC );
            // have to push back the four bytes we read for the sync
            lseek( fd, -4, SEEK_CUR );
        }
    }
    else
    {
        nextSync = lseek(fd, 0, SEEK_CUR) - 4;
        if(debug) fprintf( stderr, "found sync at %ld\n", nextSync );
    }

    prev = 0;
    lseek(fd, 0, SEEK_CUR);

    while(width--)
    {
        val = nextValue(fd, s);
#ifdef NO_EOF
      if(val == -1) return 1;
#endif
        temp = val;
        prev += temp;
        *out++ = prev;
    }
    res = s->p;

    while(s->bit > 0) nextBit(fd, s);
    while(s->p&3) nextBit(fd, s);
    return 0;
}

int pdecom(int fd, int width, int height, unsigned char *outbuf)
{
   int offset;
   int i, j, x, y, z;
   unsigned char block[2048*8], chunk[2048*8];
   unsigned char *p;
   int dointerleave = 1;
   int bytes_read = 0;

   static InputBits_t s;
   bitsInit(fd, &s);

   offset = 8*width/4;

   for(j = 0; j < height/8; j++) {

      if( lseek(fd, 0, SEEK_CUR) >= s.length )
          return bytes_read;

      p = block;
      memset( p, 0, 2048*8 ); // need to clear the block in case we can't read the full amount

      for(i = 0; i < 4; i++) {
         if (decodeX(fd, offset, p, &s) != 0)
             break; // unexpected end of file
         p += offset;
      }
      if(dointerleave) {
         unsigned char *pa = block;
         unsigned char *pb = pa + offset;
         unsigned char *pc = pb + offset;
         unsigned char *pd = pc + offset;
         z = 0;
         p = chunk;

            // a0 b0 a1 b1 a2 b2 ... aN bN
            // c0 d0 c1 d1 c2 d1 ... cN dN

         for(y = 0; y < 8; y += 2) {
            unsigned char *a = p + y*width;
            unsigned char *b = a + 1;
            unsigned char *c = a + width;
            unsigned char *d = c + 1;
            for(x = 0; x < width; x += 2) {
               a[x] = pa[z];
               b[x] = pb[z];
               c[x] = pc[z];
               d[x] = pd[z];
               z++;
            }
         }
         memcpy(outbuf, chunk, 8*width);
      } else {
         memcpy(outbuf, block, 8*width);
      }
      bytes_read += 8*width;
      outbuf += 8*width;
   }
   return bytes_read;
}
